/*
 * Copyright (c) 2022 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include <stdio.h>
#include <zephyr/kernel.h>

#include <zephyr/logging/log.h>
#include <modem/lte_lc.h>
#include <dk_buttons_and_leds.h>

#include <nrf_modem_gnss.h>

// Include SMS 
#include <modem/sms.h>

static struct nrf_modem_gnss_pvt_data_frame pvt_data;

static int64_t gnss_start_time;
static bool first_fix = false;

static K_SEM_DEFINE(lte_connected, 0, 1);
static K_SEM_DEFINE(text_ready, 0, 1);

LOG_MODULE_REGISTER(SMSwithGNSS, LOG_LEVEL_INF);
static int sms_handle;

#define SMS_STACK_SIZE 1512
#define SMS_PRIORITY 5


K_THREAD_STACK_DEFINE(sms_stack_area, SMS_STACK_SIZE);

struct k_work_q sms_work_q;
struct sms_info {
    struct k_work work;
	char number[12]; // 
    double lat;
	double lon;
} sms_work_struct;

static void lte_handler(const struct lte_lc_evt *const evt)
{
	switch (evt->type) {
	case LTE_LC_EVT_NW_REG_STATUS:
		if ((evt->nw_reg_status != LTE_LC_NW_REG_REGISTERED_HOME) &&
			(evt->nw_reg_status != LTE_LC_NW_REG_REGISTERED_ROAMING)) {
			break;
		}
		LOG_INF("Network registration status: %s",
				evt->nw_reg_status == LTE_LC_NW_REG_REGISTERED_HOME ?
				"Connected - home network" : "Connected - roaming");
		k_sem_give(&lte_connected);
		break;
	case LTE_LC_EVT_RRC_UPDATE:
		LOG_INF("RRC mode: %s", evt->rrc_mode == LTE_LC_RRC_MODE_CONNECTED ? 
				"Connected" : "Idle");
		break;	
	case LTE_LC_EVT_PSM_UPDATE:
		LOG_INF("PSM parameter update: TAU: %d, Active time: %d",
			evt->psm_cfg.tau, evt->psm_cfg.active_time);
		if (evt->psm_cfg.active_time == -1){
			LOG_ERR("Network rejected PSM parameters. Failed to setup network");
		}
		break;
	case LTE_LC_EVT_EDRX_UPDATE:
		LOG_INF("eDRX parameter update: eDRX: %f, PTW: %f",
			evt->edrx_cfg.edrx, evt->edrx_cfg.ptw);
		break;	
	default:
		break;
	}
}

static void modem_configure(void)
{
	int err;
	err = lte_lc_psm_req(true);
	if (err) {
		LOG_ERR("lte_lc_psm_req, error: %d", err);
	} 
	
	// err = lte_lc_edrx_req(true);
	// if (err) {
	// 	LOG_ERR("lte_lc_edrx_req, error: %d", err);
	// }

	LOG_INF("Connecting to LTE network");
	err = lte_lc_init_and_connect_async(lte_handler);
	if (err) {
		LOG_ERR("Modem could not be configured, error: %d", err);
		return;
	}
	k_sem_take(&lte_connected, K_FOREVER);
	LOG_INF("Connected to LTE network");
	dk_set_led_on(DK_LED2);
}

/* STEP 6 - Define a function to log fix data in a readable format */
static void print_fix_data(struct nrf_modem_gnss_pvt_data_frame *pvt_data)
{
	LOG_INF("Latitude:       %.06f", pvt_data->latitude);
	LOG_INF("Longitude:      %.06f", pvt_data->longitude);
	LOG_INF("Altitude:       %.01f m", pvt_data->altitude);
	LOG_INF("Time (UTC):     %02u:%02u:%02u.%03u",
	       pvt_data->datetime.hour,
	       pvt_data->datetime.minute,
	       pvt_data->datetime.seconds,
	       pvt_data->datetime.ms);
	// Text send after fix if it wasn't first
	if (!first_fix){
		dk_set_led_off(DK_LED1);
		dk_set_led_off(DK_LED2);
		k_msleep(100);
		// k_sem_give(&text_ready);
		// k_work_submit(&sms_work_struct.work);
	}
}


static void gnss_event_handler(int event)
{
	int err;
	switch (event) {
		
	case NRF_MODEM_GNSS_EVT_PVT:
		LOG_INF("Searching...");
		
		int num_satellites = 0;
		for (int i = 0; i < 12 ; i++) {
			if (pvt_data.sv[i].signal != 0) {
				LOG_INF("sv: %d, cn0: %d", pvt_data.sv[i].sv, pvt_data.sv[i].cn0);
				num_satellites++;
			}	
		} 
		
		err = nrf_modem_gnss_read(&pvt_data, sizeof(pvt_data), NRF_MODEM_GNSS_DATA_PVT);
		if (err) {
			LOG_ERR("nrf_modem_gnss_read failed, err %d", err);
			return;
		}
		if (pvt_data.flags & NRF_MODEM_GNSS_PVT_FLAG_FIX_VALID) {
			dk_set_led_on(DK_LED1);
			print_fix_data(&pvt_data);

			sms_work_struct.lat=pvt_data.latitude;
			sms_work_struct.lon=pvt_data.longitude;
        	k_work_submit(&sms_work_struct.work);
			if (!first_fix) {
				LOG_INF("Time to first fix: %2.1lld s", (k_uptime_get() - gnss_start_time)/1000);
				first_fix = true;
			}
			
			return;
		}
		if (pvt_data.flags & NRF_MODEM_GNSS_PVT_FLAG_SLEEP_BETWEEN_PVT){
			LOG_INF("GNSS :  At least one sleep period has passed since the last PVT notification");
			}
		if (pvt_data.flags & NRF_MODEM_GNSS_PVT_FLAG_DEADLINE_MISSED){
			LOG_INF("GNSS  'deadline missed' flag");
			}
		else if (pvt_data.flags & NRF_MODEM_GNSS_PVT_FLAG_NOT_ENOUGH_WINDOW_TIME){
			LOG_INF("Insufficient GNSS time windows");
			}
		break;
	case NRF_MODEM_GNSS_EVT_PERIODIC_WAKEUP:
		LOG_INF("GNSS has woken up");
		break;
	case NRF_MODEM_GNSS_EVT_SLEEP_AFTER_FIX:
		LOG_INF("GNSS enter sleep after fix");
		break;
	default:
		break;
	}
}

static void sms_callback(struct sms_data *const data, void *context)
{
	if (data == NULL) {
		LOG_INF("%s with NULL data\n", __func__);
		return;
	}

	if (data->type == SMS_TYPE_DELIVER) {
		/* When SMS message is received, print information */
		struct sms_deliver_header *header = &data->header.deliver;

		LOG_INF("SMS received from time:%02d-%02d-%02d %02d:%02d:%02d",
			header->time.year,
			header->time.month,
			header->time.day,
			header->time.hour,
			header->time.minute,
			header->time.second);

		LOG_INF("Text:   '%s'", data->payload);
		LOG_INF("Length: %d", data->payload_len);

		if (header->app_port.present) {
			LOG_INF("\tApplication port addressing scheme: dest_port=%d, src_port=%d\n",
				header->app_port.dest_port,
				header->app_port.src_port);
		}
		if (header->concatenated.present) {
			LOG_INF("\tConcatenated short message: ref_number=%d, msg %d/%d\n",
				header->concatenated.ref_number,
				header->concatenated.seq_number,
				header->concatenated.total_msgs);
		}
	} else if (data->type == SMS_TYPE_STATUS_REPORT) {
		LOG_INF("SMS status report received");
	} else {
		LOG_INF("SMS protocol message with unknown type received");
	}
}

static void sms_send(struct k_work *item)
{
	LOG_INF("SMS send message work");
    struct sms_info *sms_message =
        CONTAINER_OF(item, struct sms_info, work);
	char lat[10];
	char lon[10];
	char sms_text[160];
	snprintf(lat,sizeof(lat),"%f",pvt_data.latitude);
	snprintf(lon,sizeof(lon),"%f",pvt_data.longitude);
	snprintf(sms_text,sizeof(sms_text),"Fix found at %s, %s",lat,lon);
		if (sms_send_text(sms_message->number,sms_text)!=0)
			LOG_INF("Failed to SMS fix");
	LOG_INF("SMS message '%s' to number '%s'",sms_text,sms_message->number);
}

void main(void)
{

	k_work_queue_init(&sms_work_q);

	k_work_queue_start(&sms_work_q, sms_stack_area,
					K_THREAD_STACK_SIZEOF(sms_stack_area), SMS_PRIORITY,
					NULL);
	
	//Set some beginning coordinates and phone number
	strcpy(sms_work_struct.number,"+");
	sms_work_struct.lat=65.014523;
	sms_work_struct.lon=25.483436;

	k_work_init(&sms_work_struct.work, sms_send);

	if (dk_leds_init() != 0) {
		LOG_ERR("Failed to initialize the LEDs Library");
	}

	modem_configure();
	
	if (lte_lc_func_mode_set(LTE_LC_FUNC_MODE_NORMAL) != 0) {
		LOG_ERR("Failed to activate GNSS functional mode");
		return;
	}	
	
	if (nrf_modem_gnss_event_handler_set(gnss_event_handler) != 0) {
		LOG_ERR("Failed to set GNSS event handler");
		return;
	}

	if (nrf_modem_gnss_fix_interval_set(CONFIG_GNSS_PERIODIC_INTERVAL) != 0) {
		LOG_ERR("Failed to set GNSS fix interval");
		return;
	}

	if (nrf_modem_gnss_fix_retry_set(CONFIG_GNSS_PERIODIC_TIMEOUT) != 0) {
		LOG_ERR("Failed to set GNSS fix retry");
		return;
	}
	/* STEP 11 - Start the GNSS receiver*/
	LOG_INF("Starting GNSS");
	if (nrf_modem_gnss_start() != 0) {
		LOG_ERR("Failed to start GNSS");
		return;
	}	
	sms_handle = sms_register_listener(sms_callback,NULL);
	if (sms_handle<0){
		LOG_INF("SMS handle invalid");
		return;
	}
	/* STEP 12.2 - Log the current system uptime */
	gnss_start_time = k_uptime_get();
	// while(1){
	// 	if (k_sem_take(&text_ready,K_FOREVER)==0){
	// 		LOG_INF("Deactivate GNSS");
	// 		if (lte_lc_func_mode_set(LTE_LC_FUNC_MODE_DEACTIVATE_GNSS) != 0) {
	// 			LOG_ERR("Failed to activate GNSS functional mode");
	// 			return;
	// 		}
	// 		k_msleep(100);
	// 		if (sms_send_text("+","Fix found")!=0)
	// 			LOG_INF("Failed to SMS fix");
	// 		k_msleep(100);
			
	// 		if (lte_lc_func_mode_set(LTE_LC_FUNC_MODE_ACTIVATE_GNSS) != 0) {
	// 			LOG_ERR("Failed to activate GNSS functional mode");
	// 			return;
	// 		}
	// 	}
	// }
}

